/*


description
	 windowed fourier transform of data

example
 	windowed_ft <in.H par >out.H

parameters
	vfile	 	 	- velocity file input
	sx	 	float 	- source location in x
	fm		float	- fM in ricker wavelet
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <seplib.h>
#include <sepmath.h>
#include <sepfft.h>
#include <sep3dc.h>
#include "komplex.h"
#include "ccgstep.h"
#include "alloc.h"
#include "Pyramid_solver.h"
//#include "interpsinc_util2.c"
//#include <fftw3.h>
//#include <sep.main>
#define maxNdims 5
#define Ninterval 20

#define VMAX 6.22
/* F * mm ~= dd */
static complex** mm2d;          //2 dimensional model, which is in f-x domain
static complex** dd2d;          //2 dimensional data, which is in pyramid domain
static sf_double_complex* mm;          //model for convolution/deconvolution
static sf_double_complex* dd;          //model for convolution/deconvolution
static sf_double_complex* yy;          //model for convolution/deconvolution
static int twodallocated=false;   //determine if memory have been allocated for mm2d and dd2d
static int onedallocated=false;   //determine if memory have been allocated for mm and dd
int pad2(int);
float max(float,float);

struct _filehead{
  int ndims;
  int n[maxNdims];
  float o[maxNdims];
  float d[maxNdims];
};


/* F * mm ~= dd */

typedef struct _filehead filehead; 

int file_open_output(char* tag, filehead myfilehead);

void Ricker(float *array,float ot,float dt,int nt,float fm);
void sf_cconvolution(int n1,int n2,int lag,sf_complex *d1,sf_complex *d2,sf_complex *r);	
void sf_ccrosscorrelation(int n1,int n2,int lag,sf_complex *d1,sf_complex *d2,sf_complex *r);	
void sf_dcconvolution(int n1,int n2,int lag,sf_double_complex *d1,sf_double_complex *d2,sf_double_complex *r);	
void sf_dccrosscorrelation(int n1,int n2,int lag,sf_double_complex *d1,sf_double_complex *d2,sf_double_complex *r);	
void Forward_Pyramid(int adj,int add,complex **data,complex **result,int nw,int nx,float fw,float dw,float fx,float dx,int* mask); 
void Pyramid_Interp(int adj,int add,complex *data,complex *result,float dx,float dx_w,int nx,int *mask);
//void sf_Pyramid_op(int adj,int add, int nm,int nd,int nw,float fx,float fw,float dx,float dw,sf_complex **m, sf_complex **d);
void sf_Pyramid_regop(int adj,int add, int nm,int nd,int nw,float fx,float fw,float dx,float dw,sf_complex *m, sf_complex *d);
void sf_Pyramid_convolv(int adj, int add, int nm, int nd, int *mask, sf_complex *m, sf_complex *d, sf_complex *y);
void sf_Pyramid_PEFconvolv(int adj, int add, int nm, int nd, sf_complex *m, sf_complex *d, sf_complex *y);
void sf_Pyramid_MASKop(int adj, int add, int nm, int nd, int *mask, sf_complex *m, sf_complex *d,sf_complex *y);
void sf_Pyramid_LinearInterp(int adj, int add, float ox, int nx, float dx, float w, int* mask, sf_complex *m,sf_complex *d);

void sf_Pyramid_Interp1line(int adj,int add,sf_complex *model,sf_complex *data,float dx,float dx_w,int nx,int *mask);
void sf_Pyramid_AddPyramidMask(float ox, int nx, float dx, float ow, int nw, float dw, int* mask);
//void ints8c (int nxin, float dxin, float fxin, complex *yin,
//        complex yinl, complex yinr, int nxout, float *xout, complex *yout);
int main(int argc, char ** argv)
{
	int n1,n2,nx,nt,nz,maxiter,niterPT;	// n1 is nt,n2 is nx 
	float o,o1,o2,d,d1,d2,dt,dx,dz,ox,oz,ot,fpeak,fl,fh,eps,*resd,*resm;
	double re,im;
	int ndim,i,j,n,datasize,i1,i2,i3,iter,niter,*mask,*Pyramid_mask;
	float **tx_data,*txdataInterp,*temp,*temp_out;	
        float x,t,z,w,dw,pi,v,owl;
	int ix,it,iz,iw,ik,is,iwl,iwh,trunc_nw;
	complex **wx_data, **trunc_wx_data;
	sf_complex *trunc_m,*trunc_d,*model,**data,**MDdata,*data0,*PEF0,**PEF,*PEF1,*fxdataInterp;
	int fft_nw,fft_nt;
	float fft_dw,fft_dt,fft_ow;
	char temp_char[256];	
	//next two lines define parameters for output file 
	filehead outhead;
	sep3d outfile;
	int stat;	
	float max_err;
	initpar(argc, argv);
	sep_begin_prog();
	pi=3.14159265;
	/*read data from file*/
	if(0!=sep_get_number_data_axes("in",&ndim))
		seperr("unable to get dimension\n");
	for(i=1;i<=ndim;i++)
	{
		if(0!=sep_get_data_axis_par("in",&i,&n,&o,&d,temp_char))
			seperr("unable to get parameters!\n");
		if(i==1)
		{
			n1=n;
			d1=d;
			o1=o;
		}
		else
		{	
			n2=n;
			d2=d;
			o2=o;
		}
	}
	dt=d1;
	dx=d2;
	nt=n1;
	nx=n2;
	ot=o1;
	ox=o2;
	datasize=n1*n2;
	temp=(float*)alloc(sizeof(float)*datasize);
	tx_data=(float**)alloc(sizeof(float*)*nx);
	for(ix=0;ix<nx;ix++)
		tx_data[ix]=(float*)alloc(sizeof(float)*nt);

	/*put data into a 1-d arry first*/
	if(datasize*4!=sreed("in",temp,nx*nt*4))
		seperr("trouble reading data!\n");
	/*put data into 2d array*/ 
	for(ix=0;ix<nx;ix++)
		for(it=0;it<nt;it++)
			tx_data[ix][it]=temp[it+ix*nt];
	/*get alogrithm parameters */
	if(0==getch("eps","f",&eps)) eps=0;		//eps used in regularization
	if(0==getch("fl","f",&fl)) fl=5;		//lowest frequency of migration
	if(0==getch("fh","f",&fh)) fh=50;		//highest frequency of migration
	if(0==getch("niter","i",&niter)) niter=20;		//number of iteration for inversion
	if(0==getch("niterPT","i",&niterPT)) niterPT=10;	//number of iteration for pyramid transform
	        ;/*write f-x data*/

	/*determine parameter for wft and allocate space for transform*/ 	
	fft_nt=pad2(nt);			
	fft_nw=fft_nt;
	fft_ow=-pi/dt;
	fft_dw=2*pi/(dt*fft_nt);	
	fft_dt=dt;

	wx_data=(complex**)alloc(sizeof(complex*)*nx);
	for(ix=0;ix<nx;ix++)
	{
		wx_data[ix]=(complex*)alloc(sizeof(complex)*fft_nt);
		for(it=0;it<nt;it++)
		{
			wx_data[ix][it]=scmplx(tx_data[ix][it],0);
		}
		for(;it<fft_nt;it++)
		{
			wx_data[ix][it]=scmplx(0,0);
		}
	}
	

	/*doing forward fourier transform*/
	for(ix=0;ix<nx;ix++)
	{
		for(it=1;it<nt;it+=2)
			wx_data[ix][it]=scsmult(wx_data[ix][it],-1);
		cefft(wx_data[ix],fft_nt,1,1./sqrt(fft_nt));
	}
	/*truncate data such that only the part that needs to be transformed will be transformed */
	iwl=fft_nw*0.5;
	owl=fft_ow+fft_dw*iwl;
	iwh=(VMAX*pi/(2*dx)-fft_ow)/fft_dw+1;
	printf("iwl=%d, owl=%f, iwh=%d,fft_nw=%d\n",iwl, owl, iwh,fft_nw);
	if(iwh>fft_nw)
		iwh=fft_nw;
	trunc_nw=iwh-iwl;
	trunc_wx_data=(complex**)alloc(sizeof(complex*)*nx);
	trunc_m=sf_complexalloc(nx*trunc_nw);
	trunc_d=sf_complexalloc(nx*trunc_nw);
        maxiter=max(niter,niterPT);
	//printf("maxiter=%d\n",maxiter);
        temp_out=(float*)alloc(sizeof(float)*nx*trunc_nw*maxiter);
	for(ix=0;ix<nx;ix++)
	{
		trunc_wx_data[ix]=(complex*)alloc(sizeof(complex)*trunc_nw);
		for(iw=iwl;iw<iwh;iw++)
			trunc_wx_data[ix][iw-iwl]=wx_data[ix][iw];
	}
	/*write f-x data*/
	temp_out=(float*)alloc(sizeof(float)*nx*trunc_nw*maxiter);
	
	for(ix=0;ix<nx;ix++)
		for(iw=0;iw<trunc_nw;iw++)
		{
			temp_out[iw+ix*trunc_nw]=trunc_wx_data[ix][iw].re;
		}
	
	
	if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
                seperr("unable to write outfile!\n");
        outfile.o[0]=0;
        outfile.d[0]=fft_dw/2/pi;
        outfile.n[0]=trunc_nw;
        outfile.o[1]=ox;
        outfile.d[1]=dx;
        outfile.n[1]=nx;
	sep3d_set_sep3d(&outfile);
        if(0!=sep3dc_write_description("fxdata_re",&outfile))
                seperr("unable to put in parameters\n");

	if(nx*trunc_nw*4!=srite("fxdata_re",temp_out,4*nx*trunc_nw))
		seperr("trouble writing out data\n");
	
	for(ix=0;ix<nx;ix++)
		for(iw=0;iw<trunc_nw;iw++)
		{
			temp_out[iw+ix*trunc_nw]=trunc_wx_data[ix][iw].im;
		}
	
	
	if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
                seperr("unable to write outfile!\n");
        outfile.o[0]=0;
        outfile.d[0]=fft_dw/2/pi;
        outfile.n[0]=trunc_nw;
        outfile.o[1]=ox;
        outfile.d[1]=dx;
        outfile.n[1]=nx;
	sep3d_set_sep3d(&outfile);
        if(0!=sep3dc_write_description("fxdata_im",&outfile))
                seperr("unable to put in parameters\n");

	if(nx*trunc_nw*4!=srite("fxdata_im",temp_out,4*nx*trunc_nw))
		seperr("trouble writing out data\n");

	data0=sf_complexalloc(nx*trunc_nw);
	fxdataInterp=sf_complexalloc(nx*trunc_nw);
	mask=(int*)alloc(sizeof(int)*nx*trunc_nw);
	resd=(float*)alloc(sizeof(float)*niter);
	resm=(float*)alloc(sizeof(float)*niter);
	txdataInterp=(float*)alloc(sizeof(float)*nx*nt);
	for(ix=0;ix<nx;ix++)
		for(iw=0;iw<trunc_nw;iw++)
		{
			trunc_d[ix+iw*nx].r=trunc_wx_data[ix][iw].re;
			trunc_d[ix+iw*nx].i=trunc_wx_data[ix][iw].im;
			mask[ix+iw*nx]=1;
		}

	for(iw=0;iw<trunc_nw;iw++)
	{
		w=owl+iw*fft_dw;
		sf_Pyramid_MaskedTransform(1,sf_Pyramid_MASKop,sf_Pyramid_LinearInterp,sf_ccgstep,0,nx,dx,w,&trunc_m[iw*nx],&trunc_d[iw*nx],&mask[iw*nx],niterPT); 
	}
	
	for(i=0;i<nx*trunc_nw;i++)
		data0[i]=trunc_m[i];
	
	//write pyramid data
	for(ix=0;ix<nx;ix++)
		for(iw=0;iw<trunc_nw;iw++)
		{
			temp_out[iw+ix*trunc_nw]=data0[ix+iw*nx].r;
		}
	
	if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
                seperr("unable to write outfile!\n");
        outfile.o[0]=0;
        outfile.d[0]=fft_dw/2/pi;
        outfile.n[0]=trunc_nw;
        outfile.o[1]=ox;
        outfile.d[1]=dx;
        outfile.n[1]=nx;
	sep3d_set_sep3d(&outfile);
        if(0!=sep3dc_write_description("pydata_re",&outfile))
                seperr("unable to put in parameters\n");

	if(nx*trunc_nw*4!=srite("pydata_re",temp_out,4*nx*trunc_nw))
		seperr("trouble writing out data\n");
	
	//write pyramid data
	for(ix=0;ix<nx;ix++)
		for(iw=0;iw<trunc_nw;iw++)
		{
			temp_out[iw+ix*trunc_nw]=data0[ix+iw*nx].i;
		}
	
	if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
                seperr("unable to write outfile!\n");
        outfile.o[0]=0;
        outfile.d[0]=fft_dw/2/pi;
        outfile.n[0]=trunc_nw;
        outfile.o[1]=ox;
        outfile.d[1]=dx;
        outfile.n[1]=nx;
	sep3d_set_sep3d(&outfile);
        if(0!=sep3dc_write_description("pydata_im",&outfile))
                seperr("unable to put in parameters\n");

	if(nx*trunc_nw*4!=srite("pydata_im",temp_out,4*nx*trunc_nw))
		seperr("trouble writing out data\n");
	
	for(ix=0;ix<nx;ix++)
		for(iw=0;iw<trunc_nw;iw++)
		{
			//trunc_m[ix+iw*nx]=MDdata[niterMD-1][ix+iw*nx];
			trunc_m[ix+iw*nx]=data0[ix+iw*nx];
			mask[ix+iw*nx]=1;
		}

	for(iw=0;iw<trunc_nw;iw++)
	{
		w=owl+iw*fft_dw;
		sf_Pyramid_MaskedTransform(0,sf_Pyramid_MASKop,sf_Pyramid_LinearInterp,sf_ccgstep,0,nx,dx,w,&trunc_m[iw*nx],&trunc_d[iw*nx],&mask[iw*nx],niterPT); 
	}
	
	for(i=0;i<nx*trunc_nw;i++)
		fxdataInterp[i]=trunc_d[i];
	//IFFT back to t-x from f-x
	for(ix=0;ix<nx;ix++)
	{
		/*assume iwh=fft_nw*/	
		
		for(iw=iwl;iw<iwh;iw++)
		{
			wx_data[ix][iw].re=fxdataInterp[ix+(iw-iwl)*nx].r;
			wx_data[ix][iw].im=fxdataInterp[ix+(iw-iwl)*nx].i;
			//wx_data[ix][fft_nw-1-iw]=scmplx(0,0);
			wx_data[ix][fft_nw-1-iw].re=fxdataInterp[ix+(iw-iwl)*nx].r;
			wx_data[ix][fft_nw-1-iw].im=-fxdataInterp[ix+(iw-iwl)*nx].i;
		
		}
		cefft(wx_data[ix],fft_nt,-1,1./sqrt(fft_nt));
		for(it=0;it<nt;it+=2)
			txdataInterp[ix+it*nx]=wx_data[ix][it].re;
		for(it=1;it<nt;it+=2)
			txdataInterp[ix+it*nx]=-wx_data[ix][it].re;
	}
	//write reconstructed f-x data
	ndim=2;
	for(ix=0;ix<nx;ix++)
		for(iw=0;iw<trunc_nw;iw++)
		{
			temp_out[iw+ix*trunc_nw]=fxdataInterp[ix+iw*nx].r;
		}
	
	if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
                seperr("unable to write outfile!\n");
        outfile.o[0]=0;
        outfile.d[0]=fft_dw/2/pi;
        outfile.n[0]=trunc_nw;
        outfile.o[1]=ox;
        outfile.d[1]=dx;
        outfile.n[1]=nx;
	sep3d_set_sep3d(&outfile);
        if(0!=sep3dc_write_description("fxdata_reconstruct_re",&outfile))
                seperr("unable to put in parameters\n");

	if(nx*trunc_nw*4!=srite("fxdata_reconstruct_re",temp_out,4*nx*trunc_nw))
		seperr("trouble writing out data\n");
	
	
	for(ix=0;ix<nx;ix++)
		for(iw=0;iw<trunc_nw;iw++)
		{
			temp_out[iw+ix*trunc_nw]=fxdataInterp[ix+iw*nx].i;
		}
	
	if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
                seperr("unable to write outfile!\n");
        outfile.o[0]=0;
        outfile.d[0]=fft_dw/2/pi;
        outfile.n[0]=trunc_nw;
        outfile.o[1]=ox;
        outfile.d[1]=dx;
        outfile.n[1]=nx;
	sep3d_set_sep3d(&outfile);
        if(0!=sep3dc_write_description("fxdata_reconstruct_im",&outfile))
                seperr("unable to put in parameters\n");

	if(nx*trunc_nw*4!=srite("fxdata_reconstruct_im",temp_out,4*nx*trunc_nw))
		seperr("trouble writing out data\n");
	
	/*write reconstructed tx data*/
	ndim=2;
	for(ix=0;ix<nx;ix++)
		for(it=0;it<nt;it++)
		{
			temp_out[it+ix*nt]=txdataInterp[ix+it*nx];
		}
	
	if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
                seperr("unable to write outfile!\n");
        outfile.o[0]=0;
        outfile.d[0]=dt;
        outfile.n[0]=nt;
        outfile.o[1]=ox;
        outfile.d[1]=dx;
        outfile.n[1]=nx;
	sep3d_set_sep3d(&outfile);
        if(0!=sep3dc_write_description("txdata_reconstruct",&outfile))
                seperr("unable to put in parameters\n");

	if(nx*nt*4!=srite("txdata_reconstruct",temp_out,4*nx*nt))
		seperr("trouble writing out data\n");


	sep_end_prog();
	return 0;		
}

int pad2(int n)
{
        int i;
        i=1;
        while(i<n)
                i*=2;
        return i;
}

float max(float a, float b)
{
        return a>b? a:b;
}

/*generate a ricker wavelet which will be used as source in shot profile migration*/

void Ricker(float* array,float ot,float dt,int nt,float fm)
{
        float pi,t,tmp;
        int it;

        pi=3.14159265;
        for(it=0;it<nt;it++)
        {
                t=ot+it*dt;
                tmp=pi*fm*t;
                array[it]=(1-2*tmp*tmp)*exp(-tmp*tmp);
                //printf("with t=%f,array[%d]=%f\n",t,it,array[it]);
        }
}
void sf_cconvolution(int n1,int n2,int lag,sf_complex *d1,sf_complex *d2,sf_complex *r)
{
	int i;
	if(lag<n2 & lag>=0)
		for(i=0;i<n1&i<=lag;i++)	
		{
			r[0].r+=d1[i].r*d2[lag-i].r-d1[i].i*d2[lag-i].i;
			r[0].i+=d1[i].r*d2[lag-i].i+d1[i].i*d2[lag-i].r;
		}
	else if(lag>=n2 & lag<n1+n2-1)
		for(i=lag-n2+1;i<n1;i++)	
		{
			r[0].r+=d1[i].r*d2[lag-i].r-d1[i].i*d2[lag-i].i;
			r[0].i+=d1[i].r*d2[lag-i].i+d1[i].i*d2[lag-i].r;
		}
	else	
		;
}	

void sf_ccrosscorrelation(int n1,int n2,int lag,sf_complex *d1,sf_complex *d2,sf_complex *r)
{
	int i;
	if(lag<0 & lag>-n1)
		for(i=0;i<n1-lag&i<n2;i++)	
		{
			r[0].r+=d1[i-lag].r*d2[i].r+d1[i-lag].i*d2[i].i;
			r[0].i+=d1[i-lag].r*d2[i].i-d1[i-lag].i*d2[i].r;
		}
	else if(lag>=0 & lag<n2)
			for(i=lag;i<n2&i<lag+n1;i++)	
		{
			r[0].r+=d1[i-lag].r*d2[i].r+d1[i-lag].i*d2[i].i;
			r[0].i+=d1[i-lag].r*d2[i].i-d1[i-lag].i*d2[i].r;
		}
	else	
		;
		
}	

void sf_dcconvolution(int n1,int n2,int lag,sf_double_complex *d1,sf_double_complex *d2,sf_double_complex *r)
{
	int i;
	if(lag<n2 & lag>=0)
		for(i=0;i<n1&i<=lag;i++)	
		{
			r[0].r+=d1[i].r*d2[lag-i].r-d1[i].i*d2[lag-i].i;
			r[0].i+=d1[i].r*d2[lag-i].i+d1[i].i*d2[lag-i].r;
		}
	else if(lag>=n2 & lag<n1+n2-1)
		for(i=lag-n2+1;i<n1&i<=lag;i++)	
		{
			r[0].r+=d1[i].r*d2[lag-i].r-d1[i].i*d2[lag-i].i;
			r[0].i+=d1[i].r*d2[lag-i].i+d1[i].i*d2[lag-i].r;
		}
	else	
		r[0]=sf_dcmplx(0,0);
}	

void sf_dccrosscorrelation(int n1,int n2,int lag,sf_double_complex *d1,sf_double_complex *d2,sf_double_complex *r)
{
	int i;
	if(lag<0 & lag>-n1)
		for(i=0;i<-lag&i<n2;i++)	
		{
			r[0].r+=d1[n1+lag+i].r*d2[i].r+d1[n1+lag+i].i*d2[i].i;
			r[0].i+=d1[n1+lag+i].r*d2[i].i-d1[n1+lag+i].i*d2[i].r;
			}
	else if(lag>=0 & lag<n2)
		for(i=0;i<n1&i<n2-lag;i++)	
		{
			r[0].r+=d1[i].r*d2[i+lag].r+d1[i].i*d2[i+lag].i;
			r[0].i+=d1[i].r*d2[i+lag].i-d1[i].i*d2[i+lag].r;
		}
	else	
		r[0]=sf_dcmplx(0,0);
		
}	
	
void Pyramid_Interp(int adj,int add,complex *model,complex *data,float dx,float dx_w,int nx,int *mask)
{
        int ixi,ixo,ix;
        float xi,xo,x;
        if(!add)
        {
                if(adj)
                        for(ix=0;ix<nx;ix++)
                                model[ix]=scmplx(0,0);
                else
                        for(ix=0;ix<nx;ix++)
                                data[ix]=scmplx(0,0);
        }
        for(ixo=0;ixo*dx_w<(nx-1)*dx;ixo++)
        {
                x=ixo*dx_w;
                ixi=(int)(x/dx);
                xi=(x-ixi*dx)/dx;
                if(ixi<nx-1)
                {
                        if(adj)
                        {
                                        model[ixi].re+=(1.-xi)*data[ixo].re;
                                        model[ixi].im+=(1.-xi)*data[ixo].im;
                                        model[ixi+1].re+=xi*data[ixo].re;
                                        model[ixi+1].im+=xi*data[ixo].im;
                        }
                        else
                        {
                                data[ixo].re+=(1.-xi)*model[ixi].re+xi*model[ixi+1].re;
                                data[ixo].im+=(1.-xi)*model[ixi].im+xi*model[ixi+1].im;
                        }
                }
                else
                {
                        if(adj)
                                model[ixi]=data[ixo];
                        else
                                data[ixo]=model[ixi];
                }
        }
	if(!adj)
		for(;ixo<nx;ixo++)
			mask[ixo]=1;
}


void Forward_Pyramid(int adj,int add,complex **data,complex **result,int nw,int nx,float fw,float dw,float fx,float dx,int *mask)
{
        int iw,ix,ixo,nxout;
	float dx_w,w;
	float *xout;
        complex *temp_data,*temp_result;
        temp_data=(complex*)alloc(sizeof(complex)*nx);
        temp_result=(complex*)alloc(sizeof(complex)*nx);
        xout=(float*)alloc(sizeof(float)*nx);
        /*doing transform for every frequency w*/
        for(iw=0;iw<nw;iw++)
        {
		for(ix=0;ix<nx;ix++)
		{
                          xout[ix]=0;
		}
                if(adj>0)
                {
                        for(ix=0;ix<nx;ix++)
                        {
                                temp_result[ix]=result[ix][iw];
                                temp_data[ix]=scmplx(0,0);
                        }
                        /*doing forward transform using some kind of interpolation if allowable dx is bigger than current dx */
                        w=fw+iw*dw;
                        dx_w=pi*VMAX/(2*abs(w));
                        //printf("dx_w=%f\n",dx_w);
                        if(dx_w>dx)
                        {
                                Pyramid_Interp(adj,add,temp_data,temp_result,dx,dx_w,nx,&mask[iw*nx]);
                                /*
				nxout=nx;
                                for(ixo=0;ixo<nxout;ixo++)
                                       xout[ixo]=fx+ixo*dx_w;
                                //void ints8c (int nxin, float dxin, float fxin, complex yin[],complex yinl, complex yinr, int nxout, float xout[], complexyout[])
                                ints8c (nx, dx, fx, temp_data, scmplx(0,0), scmplx(0,0), nxout, xout, temp_result);
                                */
                        }
                        else
                                for(ix=0;ix<nx;ix++)
                                        temp_data[ix]=result[ix][iw];

                        /*put data to output*/
                        for(ix=0;ix<nx;ix++)
                                data[ix][iw]=temp_data[ix];
                }
                else
                {
                        for(ix=0;ix<nx;ix++)
                        {
                                temp_data[ix]=data[ix][iw];
                                temp_result[ix]=scmplx(0,0);
                        }
                        /*doing forward transform using some kind of interpolation if allowable dx is bigger than current dx */
                        w=fw+iw*dw;
			if(w==0)
				w=0.0001;
                        dx_w=pi*VMAX/(2*abs(w));
                        //printf("dx_w=%f\n",dx_w);
                        if(dx_w>dx)
                        {
                                Pyramid_Interp(adj,add,temp_data,temp_result,dx,dx_w,nx,&mask[iw*nx]);
                                /*
                                nxout=(int)((nx-1)*dx/dx_w);
				printf("nxout=%d,dx_w=%f\n",nxout,dx_w);
				for(ixo=0;ixo<nxout;ixo++)
					xout[ixo]=fx+dx_w*ixo;
				for(ixo=nxout;ixo<nx;ixo++)
					mask[ixo+iw*nx]=1;
                                //void ints8c (int nxin, float dxin, float fxin, complex yin[],complex yinl, complex yinr, int nxout, float xout[], complexyout[])	
                                ints8c (nx, dx, fx, temp_data, scmplx(0,0), scmplx(0,0), nxout, xout, temp_result);
				*/
                                
                        }
                        else
                                for(ix=0;ix<nx;ix++)
                                        temp_result[ix]=data[ix][iw];

                        /*put data to output*/
                        for(ix=0;ix<nx;ix++)
                                result[ix][iw]=temp_result[ix];
                }
        }
	//free(temp_data);
	//free(temp_result);
	//free(xout);
}

void sf_Pyramid_convolv(int adj, int add, int nm, int nd, int *mask, sf_complex *m, sf_complex *d,sf_complex *y)
{
	int fft_n,i;
	if(!onedallocated)
	{
		dd=(sf_double_complex*)alloc(sizeof(sf_double_complex)*nd*2); 
		yy=(sf_double_complex*)alloc(sizeof(sf_double_complex)*nd*2); 
		mm=(sf_double_complex*)alloc(sizeof(sf_double_complex)*nd*2);
		//printf("nd=%d\n",nd);

		onedallocated=true;
	}
	if(adj)
	{
		for(i=0;i<nd;i++)
		{
			yy[i]=sf_dcmplx(y[i].r,y[i].i);
			if(mask[i+1]==0)
				dd[i]=sf_dcmplx(0,0);
			else
				dd[i]=sf_dcmplx(d[i].r,d[i].i);
		}
		if(!add)
			for(i=0;i<nm;i++)
				m[i]=sf_cmplx(0,0);
		for(i=0;i<nm;i++)
			mm[i]=sf_dcmplx(0,0);
		
		for(i=0;i<nm;i++)
		{
			sf_dccrosscorrelation(nd,nd,i,yy,dd,&mm[i]);
			m[i].r+=mm[i].r;
			m[i].i+=mm[i].i;
		}
	}
	else
	{
		if(!add)
			for(i=0;i<nd;i++)
				d[i]=sf_cmplx(0,0);
		for(i=0;i<nd;i++)
			dd[i]=sf_dcmplx(0,0);
		for(i=0;i<nd;i++)
			yy[i]=sf_dcmplx(y[i].r,y[i].i);
		for(i=0;i<nm;i++)
			mm[i]=sf_dcmplx(m[i].r,m[i].i);
		for(i=0;i<nd;i++)
		{
			if(mask[i+1]==1)
			{
				sf_dcconvolution(nm,nd,i,mm,yy,&dd[i]);
				d[i].r+=dd[i].r;
				d[i].i+=dd[i].i;
			}
		}
	}
}

void sf_Pyramid_MASKop(int adj, int add, int nm, int nd, int *mask, sf_complex *m, sf_complex *d,sf_complex *y)
{
	int fft_n,i;
	if(adj)
	{
		if(!add)
			for(i=0;i<nm;i++)
				d[i]=sf_cmplx(0,0);
		
		for(i=0;i<nm;i++)
		{
			if(mask[i]>0)
			{	
				d[i].r+=y[i].r;
				d[i].i+=y[i].i;
			}
		}
	}
	else
	{
		if(!add)
			for(i=0;i<nd;i++)
				y[i]=sf_cmplx(0,0);
		for(i=0;i<nd;i++)
		{
			if(mask[i]>0)
			{
				y[i].r+=d[i].r;
				y[i].i+=d[i].i;
			}
		}
	}
}

void sf_Pyramid_PEFconvolv(int adj, int add, int nm, int nd, sf_complex *m, sf_complex *d,sf_complex *y)
{
	int fft_n,i;
	if(!onedallocated)
	{
		dd=(sf_double_complex*)alloc(sizeof(sf_double_complex)*nd*2); 
		yy=(sf_double_complex*)alloc(sizeof(sf_double_complex)*nd*2); 
		mm=(sf_double_complex*)alloc(sizeof(sf_double_complex)*nd*2);
		onedallocated=true;
	}
	if(adj)
	{	
		if(!add)
			for(i=0;i<nd;i++)
				d[i]=sf_cmplx(0,0);
		for(i=0;i<nd;i++)
			dd[i]=sf_dcmplx(0,0);
		for(i=0;i<nd;i++)
			yy[i]=sf_dcmplx(y[i].r,y[i].i);
		for(i=0;i<nm;i++)
			mm[i]=sf_dcmplx(m[i].r,m[i].i);
		for(i=0;i<nd;i++)
		{
			sf_dccrosscorrelation(nm,nd,i+1,mm,yy,&dd[i]);
			d[i].r+=dd[i].r+y[i].r;
			d[i].i+=dd[i].i+y[i].i;
		}
	}
	else
	{
		if(!add)
			for(i=0;i<nd;i++)
				y[i]=sf_cmplx(0,0);
		for(i=0;i<nd;i++)
			yy[i]=sf_dcmplx(0,0);
		for(i=0;i<nd;i++)
			dd[i]=sf_dcmplx(d[i].r,d[i].i);
		for(i=0;i<nm;i++)
			mm[i]=sf_dcmplx(m[i].r,m[i].i);
		for(i=0;i<nd;i++)
		{
			sf_dcconvolution(nm,nd,i-1,mm,dd,&yy[i]);
			y[i].r+=yy[i].r+d[i].r;
			y[i].i+=yy[i].i+d[i].i;
		}
	}
}

/*linear operator used for Pyramid transform, going forth and back*/
void sf_Pyramid_LinearInterp(int adj, int add, float ox, int nx, float dx, float w, int* mask, sf_complex *m,sf_complex *d)
{
	float pi,dx_w;
	pi=3.14159265;
	if(w==0)
		w=1;	
       	dx_w=pi*VMAX/(2*abs(w));
	sf_Pyramid_Interp1line(adj,add,m,d,dx,dx_w,nx,mask);
}


void sf_Pyramid_Interp1line(int adj,int add,sf_complex *model,sf_complex *data,float dx,float dx_w,int nx,int *mask)
{
        int ixi,ixo,ix;
        float xi,xo,x;
        if(!add)
        {
                if(adj)
                        for(ix=0;ix<nx;ix++)
                                model[ix]=sf_cmplx(0,0);
                else
                        for(ix=0;ix<nx;ix++)
                                data[ix]=sf_cmplx(0,0);
        }
        for(ixo=0;ixo<nx;ixo++)
        {
                x=ixo*dx;
                ixi=(int)(x/dx_w);
                xi=(x-ixi*dx_w)/dx_w;
		
                if(ixi<nx-1)
                {
                        if(adj)
                        {
                                        model[ixi].r+=(1.-xi)*data[ixo].r;
                                        model[ixi].i+=(1.-xi)*data[ixo].i;
                                        model[ixi+1].r+=xi*data[ixo].r;
                                        model[ixi+1].i+=xi*data[ixo].i;
                        }
                        else
                        {
                                data[ixo].r+=(1.-xi)*model[ixi].r+xi*model[ixi+1].r;
                                data[ixo].i+=(1.-xi)*model[ixi].i+xi*model[ixi+1].i;
                        }
                }
		
                else
                {
                        if(adj)
                        {
                                        model[ixi].r+=(1.-xi)*data[ixo].r;
                                        model[ixi].i+=(1.-xi)*data[ixo].i;
                        }
                        else
                        {
                                data[ixo].r+=(1.-xi)*model[ixi].r;
                                data[ixo].i+=(1.-xi)*model[ixi].i;
                        }
                }
        }
	/*
	if(!adj)
		for(ixo=(nx-1)*dx/dx_w+1;ixo<nx;ixo++)
			mask[ixo]=1;
	*/
}

/*linear operator used for Pyramid transform, going forth and back*/
void sf_Pyramid_AddPyramidMask(float ox, int nx, float dx, float ow, int nw, float dw, int* mask)
{
	float pi,dx_w,w;
	int iw,ix;
	pi=3.14159265;
	for(iw=0;iw<nw;iw++)
	{
		w=ow+iw*dw;
		if(w==0)
			w+=0.1;	
       		dx_w=pi*VMAX/(2*abs(w));
		for(ix=(nx-1)*dx/dx_w+1;ix<nx;ix++)
			mask[ix+iw*nx]=1;
	}
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
